from __future__ import annotations

# fmt:off
import aim

from aim.sdk.repo import Repo
from aim.sdk.run import get_installed_packages, get_environment_variables, get_git_info, sys
from aim.sdk.run import Optional, BasicRunAutoClean, TreeView, logger, backup_run, RunStatusReporter, LocalFileManager, pathlib, RemoteFileManager, ScheduledStatusReporter, RemoteRunHeartbeatReporter, RunTracker
from aim.sdk.base_run import get_repo, analytics, MissingRunError, Dict
from aim.ext.resource import ResourceTracker, DEFAULT_SYSTEM_TRACKING_INT

from t3w.utils.misc import generate_run_hash as t3w_generate_run_hash


# [feat] Allow users to specify the run_hash: https://github.com/aimhubio/aim/pull/2738/commits/fd48e0a3dc7e505fff86791dd30d70b3d899b008
class Run(aim.Run):

    def __init__(self, run_hash: str | None = None, *, repo: str | Repo | pathlib.Path | None = None, read_only: bool = False, experiment: str | None = None, force_resume: bool = False, system_tracking_interval: int | float | None = 600, log_system_params: bool | None = False, capture_terminal_logs: bool | None = True):

        self._resources: Optional[BasicRunAutoClean] = None

        #########################
        # aim.BaseRun.__init__  #
        #########################

        self._hash = None
        self._lock = None

        self.read_only = read_only
        self.repo = get_repo(repo)
        if self.read_only:
            assert run_hash is not None
            self.hash = run_hash
        else:
            if run_hash is None or not self.repo.run_exists(run_hash):
                self.hash = run_hash or t3w_generate_run_hash()
            # if run_hash is None:
            #     self.hash = generate_run_hash()
                analytics.track_event(event_name='[Run] Create new run')
            else:
            # elif self.repo.run_exists(run_hash):
                self.hash = run_hash
                analytics.track_event(event_name='[Run] Resume run')
            # else:
            #     raise MissingRunError(f'Cannot find Run {run_hash} in aim Repo {self.repo.path}.')
            self._lock = self.repo.request_run_lock(self.hash)
            self._lock.lock(force=force_resume)

        self.meta_tree: TreeView = self.repo.request_tree(
            'meta', self.hash, read_only=read_only, from_union=True
        ).subtree('meta')
        self.meta_run_tree: TreeView = self.meta_tree.subtree('chunks').subtree(self.hash)

        self._series_run_trees: Dict[int, TreeView] = None

        #########################
        # aim.BasicRun.__init__ #
        #########################

        self.meta_attrs_tree: TreeView = self.meta_tree.subtree('attrs')
        self.meta_run_attrs_tree: TreeView = self.meta_run_tree.subtree('attrs')

        if not read_only:
            logger.debug(f'Opening Run {self.hash} in write mode')

            if self.check_metrics_version():
                if self.repo.is_remote_repo:
                    logger.warning(f'Cannot track Run with remote repo {self.repo.path}. Please upgrade repo first '
                                   f'with the following command:')
                    logger.warning(f'aim storage --repo {self.repo.path} upgrade 3.11+ \'*\'')
                    raise RuntimeError
                else:
                    logger.warning(f'Detected sub-optimal format metrics for Run {self.hash}. Upgrading...')
                    backup_path = backup_run(self)
                    try:
                        self.update_metrics()
                        logger.warning(f'Successfully converted Run {self.hash}')
                        logger.warning(f'Run backup can be found at {backup_path}. '
                                       f'In case of any issues the following command can be used to restore data: '
                                       f'`aim storage --repo {self.repo.root_path} restore {self.hash}`')
                    except Exception as e:
                        logger.error(f'Failed to convert metrics. {e}')
                        logger.warning(f'Run backup can be found at {backup_path}. '
                                       f'To restore data please run the following command: '
                                       f'`aim storage --repo {self.repo.root_path} restore {self.hash}`')
                        raise

        self._props = None
        self._checkins = None
        self._heartbeat = None

        if not read_only:
            if not self.repo.is_remote_repo:
                self._checkins = RunStatusReporter(self.hash, LocalFileManager(self.repo.path))
                progress_flag_path = pathlib.Path(self.repo.path) / 'meta' / 'progress' / self.hash
                self._heartbeat = ScheduledStatusReporter(self._checkins, touch_path=progress_flag_path)
            else:
                self._checkins = RunStatusReporter(self.hash, RemoteFileManager(self.repo._client, self.hash))
                self._heartbeat = RemoteRunHeartbeatReporter(self.repo._client, self.hash)

            try:
                self.meta_run_attrs_tree.first_key()
            except (KeyError, StopIteration):
                # no run params are set. use empty dict
                self[...] = {}
            self.meta_run_tree['end_time'] = None
            self.props
        if experiment:
            self.experiment = experiment

        self._tracker = RunTracker(self)
        self._resources = BasicRunAutoClean(self)

        #####################
        # aim.Run.__init__  #
        #####################

        self._system_resource_tracker: ResourceTracker = None
        if not read_only:
            if log_system_params:
                self['__system_params'] = {
                    'packages': get_installed_packages(),
                    'env_variables': get_environment_variables(),
                    'git_info': get_git_info(),
                    'executable': sys.executable,
                    'arguments': sys.argv
                }

            if ResourceTracker.check_interval(system_tracking_interval) or capture_terminal_logs:
                current_logs = self.get_terminal_logs()
                log_offset = current_logs.last_step() + 1 if current_logs else 0
                self._system_resource_tracker = ResourceTracker(self._tracker,
                                                                system_tracking_interval,
                                                                capture_terminal_logs,
                                                                log_offset)
                self._system_resource_tracker.start()
                self._resources.add_extra_resource(self._system_resource_tracker)

# fmt:on



from pathlib import Path
import mgzip
import pickle


class CachedDataset:

    def __init__(self, raw_dataset, cache_dir, compresslevel=1):
        self.raw_dataset = raw_dataset
        self.cache_dir = Path(cache_dir)
        self.compresslevel = compresslevel
        Path(self.cache_dir).mkdir(exist_ok=True)

    def __len__(self):
        return len(self.raw_dataset)

    def __getitem__(self, index):
        cached_datum = self.cache_dir / (str(index) + ".gz")
        if cached_datum.exists():
            with mgzip.open(str(cached_datum)) as fp:
                datum = pickle.load(fp)
        else:
            datum = self.raw_dataset[index]
            with mgzip.open(str(cached_datum), "wb", compresslevel=self.compresslevel) as fp:
                pickle.dump(datum, fp)
        return datum

    @property
    def datum_type(self):
        return self.raw_dataset.datum_type



import torch, os, math
import numpy as np

from typing import List
from torch import nn, Tensor


suppress_traceback = []


# Some monkey-patch for more concise output.
if not os.environ.get("T3W_VERBOSE", False):

    def _ndarray_summary(self:np.ndarray):
        if len(self.shape):
            nan_count = np.isnan(self).sum() if isinstance(self, np.floating) else 0
            nan_str = f" nan={nan_count}," if nan_count else ""
            return f"NDArray({tuple(self.shape)}, {str(self.dtype).replace('torch.', '')},{nan_str} \"cpu\")"
        else:
            return f"Scalar({self.item()}, {str(self.dtype).replace('torch.', '')}, \"cpu\")"

    def _tensor_summary(self:Tensor):
        if len(self.shape):
            nan_count = torch.isnan(self).sum() if torch.is_floating_point(self) else 0
            nan_str = f" nan={nan_count}," if nan_count else ""
            return f"Tensor({tuple(self.shape)}, {str(self.dtype).replace('torch.', '')},{nan_str} \"{str(self.device)}\")"
        else:
            return f"Scalar({self.item()}, {str(self.dtype).replace('torch.', '')}, \"{str(self.device)}\")"

    def _module_summary(self:nn.Module):
        nparams = sum(p.numel() for p in self.parameters())
        try: device = next(self.parameters()).device
        except StopIteration: device = None
        head_repr = self.extra_repr()
        tail_repr = f"nparams={millify(nparams)}, device=\"{device}\""
        args_repr = f"{head_repr}, {tail_repr}" if head_repr else tail_repr
        return f"{self.__class__.__name__}({args_repr})"

    def _verbose_ndarray(x):
        np.set_string_function(None)
        x = repr(x)
        np.set_string_function(_ndarray_summary, repr=True)
        return x

    _module_raw_repr = nn.Module.__repr__

    def _verbose_module(x:nn.Module):
        nn.Module.__repr__ = _module_raw_repr
        x = repr(x)
        nn.Module.__repr__ = _module_summary
        return x

    _VERBOSE_REPRS = {
        Tensor: Tensor.__repr__,
        nn.Module: _verbose_module,
        np.ndarray: _verbose_ndarray,
        list: lambda c: [verbose(i) for i in c],
        tuple: lambda c: tuple([verbose(i) for i in c]),
        dict: lambda c: {k:verbose(v) for k, v in c.items()},
    }

    class ReprStr(str):
        def __repr__(self):
            return self

    def verbose(data):
        for T, fn in _VERBOSE_REPRS.items():
            if isinstance(data, T):
                return ReprStr(fn(data))
        return data

    Tensor.__repr__ = _tensor_summary
    nn.Module.__repr__ = _module_summary
    np.set_string_function(_ndarray_summary, repr=True)

    try:

        # suppress_traceback.append(__file__)
        suppress_traceback.append(torch)

        import rich.traceback
        _old_from_exception = rich.traceback.Traceback.from_exception
        def _rich_traceback_from_exception(cls, *args, suppress:List=[], **kwargs):
            suppress.extend(suppress_traceback)
            kwargs['locals_max_length'] = os.environ.get("T3W_LOCALS_MAXLEN", 3)
            return _old_from_exception(*args, suppress=suppress, **kwargs)
        rich.traceback.Traceback.from_exception = classmethod(_rich_traceback_from_exception)
    except ImportError: pass

else:
    def verbose(data):
        return data


def millify(n:int, names=["", "K", "M", "B", "T"]):
    """format an integer number into a human-friendly string.

    Args:
        n (int): the number to be formatted.
        names (list, optional): postfix string for one, thousand, million, etc. Defaults to ["", " K", " M", " B", " T"].

    Returns:
        str: the fomatted string.
    """
    n = float(n)
    millidx = max(0, min(len(names) - 1, int(math.floor(0 if n == 0 else math.log10(abs(n)) / 3))))
    return "{:.1f}{}".format(n / 10 ** (3 * millidx), names[millidx]).replace(".0", "")



"""
Copyright (C) 2023 Yuyao Huang - All Rights Reserved

You may use, distribute and modify this code under the terms of the Apache 2.0
license, which unfortunately won't be written for another century. You should
have received a copy of the Apache 2.0 license with this file. If not, please
write to: huangyuyao@outlook.com, or visit:
https://github.com/tjyuyao/t3w/blob/main/LICENSE
"""

# docstring style reference: https://www.sphinx-doc.org/en/master/usage/extensions/example_google.html
# reStructuredText Primer: https://www.sphinx-doc.org/en/master/usage/restructuredtext/basics.html
# reStructuredText Python: https://www.sphinx-doc.org/en/master/usage/restructuredtext/domains.html#cross-referencing-python-objects

from __future__ import annotations
from abc import ABC, abstractmethod
import torch, gzip, random, os, weakref
import numpy as np
from typing import Any, Sequence, NewType, Optional, Mapping, Type, Hashable, Callable, Dict, Literal, Union, List
from enum import Enum
from functools import wraps
from glob import glob
from torch.utils.data import DataLoader, Sampler, default_collate, DistributedSampler
from torch import nn, Tensor, FloatTensor
from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler
from torch.nn.parallel import DistributedDataParallel
from jaxtyping import Float, Shaped, Bool
from dataclasses import dataclass, asdict
from pathlib import Path
from rich.pretty import pprint
import torch.distributed as dist
import torch.multiprocessing as mp

from .utils.verbose import verbose


def manual_seed(seed:int, strict:bool=False):
    """set all default random generators seed to ``seed``.

    Args:
        seed (int): the value to be set.
        strict (bool, optional): whether to use deterministic pytorch algorithms for better reproducibility. Defaults to False.
    """
    torch.manual_seed(seed)
    random.seed(seed)
    np.random.seed(seed)
    if strict:
        os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8"
        torch.use_deterministic_algorithms(True, warn_only=True)


FloatScalarTensor = NewType("MiniBatchFloats", Float[Tensor, ""])
MiniBatchFloats = NewType("MiniBatchFloats", Float[Tensor, "minibatch"])
"""This is a sequence of float, the return type for :meth:`IMiniBatchMetric.forward` method, each scalar float value corresponds to an example in the minibatch.
"""

StepReturnDict = NewType("StepReturnDict", Dict[Literal["losses", "metrics"], Dict[str, MiniBatchFloats]])
"""This is the returned type of :meth:`TopLevelModule.step`. Users may can receive data of this type in the event :meth:`ISideEffect.on_train_step_finished`.
"""


class Interface(ABC):
    """This is a special abstract class to indicate its direct subclasses are interface and should be implemented by the user.

    Note:
        Every direct subclass of :class:`Interface` has its name starting with letter "I".
    """
    pass


class IDatum(Interface):
    """The interface for a single datum. Your dataset is meant to have abundant of this type of objects.

    Note:
        :class:`TrainLoop` or :class:`EvalLoop` can directly make use of attributes and methods defined here through the
        :attr:`IDataset.datum_type` link.
    """

    uuid: Optional[Hashable]
    """universal unique identifier for tracking datum information, optional."""

    train_batch_size: int = 2
    """:class:`TrainLoop` will use this value through :attr:`IDataset.datum_type` as default batch size value."""

    val_batch_size: int = 2
    """:class:`EvalLoop` will use this value through :attr:`IDataset.datum_type` as default batch size value."""

    num_workers: int = 0
    """:class:`TrainLoop` or :class:`EvalLoop` will use this value through :attr:`IDataset.datum_type` for `DataLoader.num_workers`."""

    @staticmethod
    def collate(data: Sequence["IDatum"]) -> "IMiniBatch":
        """collate a mini batch of data into an :class:`IMiniBatch` consumable by user's model and :class:`IMiniBatchMetric`.

        This function is called by :class:`DataLoader` in the :class:`TrainLoop` or :class:`EvalLoop` to collate independently sampled data
        into a mini batch, before which is passed to the :meth:`TopLevelModule.forward`.

        Note:
            It is strongly suggested not to use the pytorch's :func:`default_collate` function but override it here.
            This is because :func:`default_collate` does not support extra types other than numerical tensors, and make
            users not easy to manage their batch data in an object-oriented-programming paradigm. Be explicit, and
            write your own collate function to produce your :class:`IMiniBatch` type.

        Args:
            data (list(IDatum)): sampled mini batch of data as a sequence.

        Returns:
            IMiniBatch: the collated mini batch.
        """
        return default_collate(data)

    @classmethod
    def from_uuid(cls, uuid: Hashable) -> "IDatum":
        """build the exact datum from given ``uuid``.

        Ignore this factory method for basic training workflow. It is
        useful for some workflows that mine, track and analyse specific interesting data.
        One may embed dataset_type, split, and index information to fully recover a datum's identity.

        Args:
            uuid (Hashable): universal unique identifier of the disired datum.

        Returns:
            IDatum:
        """
        dataset_type, root, split, index = uuid
        return dataset_type(root, split)[index]

    def __repr__(self) -> str:
        C1 = self.__class__.__name__
        D1 = self.uuid[0].__name__
        return f"{C1}.from_uuid(({D1}, {', '.join([repr(x) for x in self.uuid[1:]])}))"


class IMiniBatch(Interface):
    """The interface for a minibatch of datum.

    It plays the central role of a data context passing around models, metrics, losses and callbacks.
    All interfaces in ``t3w`` has a strict typing requirement, this especially emphasize that functions
    are required to return specific type of data, instead of dynamic types. The flexibility of ``t3w``
    is not harmed though, because it is user's freedom to modify the definition of subclasses of the
    :class:`IMiniBatch` interface. And all the functions receives a :class:`IMiniBatch` as input, are
    allowed (or supposed) to modify it in place. Therefore, standard behaviors of ``t3w`` core libraries
    and :class:``ISideEffect`` based plugin system can rely on the typing system, while users can write
    flexible code in their independent namespace.

    See also:
      * :meth:`TopLevelModule.forward`
      * :meth:`IMiniBatchMetric.forward`
      * :meth:`ISideEffect.on_eval_step_started`
      * :meth:`ISideEffect.on_eval_step_finished`
      * :meth:`ISideEffect.on_train_step_started`
      * :meth:`ISideEffect.on_train_step_finished`
    """

    # @cached_property
    # def batch_size(self) -> int:
    #     raise AttributeError(f"'{self.__class__.__name__}' object has no attribute 'batch_size', it should be implemented by user in the \"collate\" function.")

    _nopad_batch_size: int = None
    """

    Note:
        In DDP mode, when data cannot be evenly distributed to multiple devices at the last iteration of a dataset epoch, :class:`DistributedSampler` will pad the samples to ensure same batch size on all devices. The padded samples would cause minor error in the evaluation result and should be avoided. Fortunately, :class:`EvalLoop` automatically mantains ``_nopad_batch_size``, so :meth:`IMiniBatchMetric.__call__` can handle this case properly during computing an average.

    See also:
      :meth:`EvalLoop.mark_padding`
    """

    model:"TopLevelModule" = None
    """The :class:`TopLevelModule` which is consuming and modifying this instance of :class:`IMiniBatch`.

    The :class:`TopLevelModule` will fill this attribute right before calling the `forward()` method of its internal user_model. Therefore, the ``user_model.forward()``, :meth:`ISideEffect.on_evak_step_finished`, and :meth:`ISideEffect.on_train_step_finished` can make use of it. Before that, this attribute is defaulted to ``None``.
    """

    def to(self, device: torch.device):
        """defines how to move this mini-batch type into specified device.

        The default implementation in the base class :class:`IMiniBatch` moves all direct attributes which are instances of ``torch.Tensor`` to target device. This can be good enough for common usage, but nothing stops you from customizing it.

        Note:
            :meth:`TopLevelModule.forward` will call this function right before calling ``user_model.forward()``, so don't bother to do it yourself. You only need to specify the target device at the :class:`TopLevelModule` level using its :meth:`TopLevelModule.to` method.

        See also:
            :meth:`TopLevelModule.to`

        Args:
            device (torch.device): target device.
        """
        for name in self.__dict__:
            attr = getattr(self, name, None)
            if isinstance(attr, Tensor):
                setattr(self, name, attr.to(device))
        return self


class IDataset(Interface):
    """The interface for the entire dataset (and its split).

    This is very simillar to ``torch.Dataset`` (sized dataset) interface, with only subtle modifications
    that :meth:`__getitem__` is required to return an instance of :attr:`datum_type`.
    """

    datum_type: Type[IDatum]
    
    @property
    @abstractmethod
    def datum_type(self) -> Type[IDatum]:
        """User implemented IDatum subclass' **typename**.

        Note:
            Subclass of ``IDataset`` must specify this attribute in order to fetch the class attribute including :attr:`IDatum.train_batch_size`, :attr:`IDatum.val_batch_size`, and :attr:`IDatum.num_workers`, etc.
        """
        pass

    def __init__(self, root:str, split:str=None) -> None:
        """

        Args:
            root (str): path of the root directory of specified dataset.
            split (str, optional): identifier for a subset. Defaults to None.
        """
        self.root = root
        self.split = split

    def __len__(self) -> int:
        """

        Returns:
            int: the number of data in the current split of dataset.
        """
        pass

    def __getitem__(self, index: int) -> IDatum:
        """_summary_

        Args:
            index (int): _description_

        Returns:
            IDatum: the ``index``-th datum in the dataset.
        """
        pass


class IMiniBatchMetric(nn.Module, Interface):
    """The interface of compute metric for datum in a mini-batch.

    Note:
        We differentiate the :class:`IMiniBatchMetric` and :class:`IDatasetMetric`,
        where the former compute metric value for a batch of data, while the
        latter aggregate datum metric of each batch for a entire dataset (typically
        an "average meter"). The dataset level metric mainly focus on correct computation
        and synchronization of variable batch sizes and among devices.

    See also:
        :class:`ILoss`, :class:`IDatasetMetric`, :class:`AverageMetric`.
    """

    higher_better: bool
    """specifies whether higher value of the metric implies better performance.
    This can be useful for e.g. metric based best model saving. Always
    explicitly specify this class variable in your subclass definition.
    """

    def forward(self, mb: IMiniBatch) -> MiniBatchFloats | FloatScalarTensor:
        """

        Calling of this method is delegated to :class:`TrainLoop` or :class:EvalLoop` at their construction time.
        The ``mb`` argument must have been through the user_model's ``forward`` method already, and the loops pass
        it on to metrics to calculate the metric value, which must be a float scalar tensor.

        Args:
            mb (IMiniBatch): the mini-batch of data which have been processed by user_model.

        Returns:
            MiniBatchFloats | FloatScalarTensor: returns a sequence or a single value (regarded as the same for every example in the minibatch) of metric values.
        """
        pass

    def __call__(self, mb: IMiniBatch) -> FloatScalarTensor:
        """ call forward() and compute the mean value of it.
        """
        metric_values:FloatTensor = super().__call__(mb).float()
        if metric_values.numel() == 1:
            return metric_values
        else:
            return metric_values[:mb._nopad_batch_size].mean()


class ILoss(IMiniBatchMetric):
    """The interface of compute loss for datum in a mini-batch.

    Note:
        In t3w, we adopt the fact that a loss function is a special type of metric that support backpropagation. Therefore :class:`ILoss` inherites :class:`IMiniBatchMetric` and you can use an ``ILoss`` whereever an ``IMiniBatchMetric`` is suited.
    """

    loss_reweight: float = 1.
    """Losses has the raw version and reweighted version in lots of circumstances. Use this standard attribute to specify your weight of current loss.

    Note:
        In the "losses" subdict of :class:`StepReturn`, the value of each loss will be reported as a pair of float value ``(loss_reweight, loss_raw_value)`` with event :meth:`on_train_step_finished` emitted. This is the standard behavior that extension codes can rely on. Non-loss metrics are reported as pure floats.
    """

    higher_better: bool = False
    """higher value of a loss always implies worse performance. Don't bother to specify it in the subclasses."""

    def forward(self, mb: IMiniBatch) -> MiniBatchFloats | FloatScalarTensor:
        """

        Calling of this method is delegated to :class:`TrainLoop` at its construction time.
        The ``mb`` argument must have been through the user_model's ``forward`` method already, and the loops pass
        it on to various losses to calculate the loss value, which must be a float scalar tensor. And
        then the losses will be reweighted and summed together for a backward autodiff pass.

        Args:
            mb (IMiniBatch): the mini-batch of data which have been processed by user_model.

        Returns:
            FloatScalarTensor: return loss scalar value.
        """
        pass


class LearningRate(IMiniBatchMetric):
    """This class report current learning rate through the standard metric interface.

    This is not a typical metric but it is a commonly used one agnostic to tasks. So we implement it early here.
    It is also a good demonstration of how to use the exposed :class:`TopLevelModule` as the :attr:`IMiniBatch.model` attribute. Since the metric computation is applied after the user_model's ``forward()``, the ``model`` attribute is absolutely available in a :meth:`IMiniBatchMetric.forward` method.
    """

    def __init__(self, param_group=0) -> None:
        """

        Args:
            param_group (int, optional): to show learning rate of which. Defaults to 0.
        """
        super().__init__()
        self.param_group = param_group

    def forward(self, mb: IMiniBatch) -> FloatScalarTensor:
        """

        Args:
            mb (IMiniBatch): a mini-batch during training.

        Returns:
            FloatScalarTensor: the learning rate of self.param_group
        """
        return torch.tensor(mb.model.optim.param_groups[self.param_group]['lr'])


class IDatasetMetric(IMiniBatchMetric):
    """The interface of a dataset level metric (aggregation algorithm on multi devices).

    Note:
        We differentiate the :class:`IMiniBatchMetric` and :class:`IDatasetMetric`,
        where the former compute metric value for a batch of data, while the
        latter aggregate datum metric of each batch for a entire dataset
        (typically an "average meter"). The dataset level metric mainly focus on
        correct computation and synchronization of variable batch sizes and
        among devices.

    Warning:
        Typically, :class:`TrainLoop` only accepts datum metric because of
        changing parameters, while :class:`EvalLoop` only accepts dataset metric
        because it want statistics on the entire dataset, and only call
        :meth:`IDatasetMetric.synchronize` once right before emitting event
        :meth:`on_eval_finished`. Other use cases are still possible and
        ``synchronize()`` should also remain correct after arbitrary times of
        call, e.g. a running average dataset metric can be used in a train loop,
        but it is still considered non-standard behavior in ``t3w`` and should
        prefer implemented and maintained in user space using the side effects
        system.

    See also:
        :class:`IMiniBatchMetric`, :class:`AverageMetric`.
    """

    minibatch_metric: IMiniBatchMetric
    """The dataset metric has a standard behavior to composite a datum metric instance, and the calling
    of the datum metric is delegated to :meth:`update`.
    """

    def __init__(self, minibatch_metric: IMiniBatchMetric) -> None:
        """store ``minibatch_metric`` and reset the statistics.

        Args:
            minibatch_metric (IMiniBatchMetric): internal minibatch_metric instance to embed.
        """
        super().__init__()
        self.minibatch_metric = minibatch_metric
        self.reset()

    def forward(self, mb: IMiniBatch) -> MiniBatchFloats:
        """allows using dataset metric like datum metric.

        Warning:
            It is non-standard behavior so please be careful implementing a
            :class:`IDatasetMetric` that supposed to be used as a datum metric,
            especially ensure multiple call of synchronization among multiple
            devices does not have ambiguity, for example by separately store the
            synchronized and local version of states.

            **This will be changed in a future version, by using a breaking API with stateless synchronization.**

        Args:
            mb (IMiniBatch): a mini-batch of data.

        Returns:
            MiniBatchFloats: current dataset metric value.
        """
        self.update(mb)
        self.synchronize()
        return torch.tensor(self.eval())

    def eval(self) -> float:
        """get target metric value based on internal statistics."""
        pass

    def reset(self) -> None:
        """clear internal statistics."""
        pass

    def update(self, mb: IMiniBatch) -> None:
        """clear internal statistics."""
        pass

    def synchronize(self) -> None:
        """synchronize local statistics."""
        pass

    def float(self):
        return self.eval()

    # def __repr__(self):
    #     return str(self.eval())


class AverageMetric(IDatasetMetric):
    """The average metric across data.
    """

    def eval(self) -> float:
        return self.sum / max(self.cnt, 1)

    def reset(self) -> None:
        self.sum = 0.
        self.cnt = 0

    def update(self, mb: IMiniBatch) -> None:
        new_val = self.minibatch_metric(mb).item()
        k = mb._nopad_batch_size
        self.sum += new_val * k
        self.cnt += k

    def synchronize(self) -> None:
        if not dist.is_initialized(): return
        object_list = [None for _ in range(dist.get_world_size())]
        dist.all_gather_object(
            object_list=object_list,
            obj=dict(sum=self.sum, cnt=self.cnt)
        )
        if dist.get_rank() == 0:
            self.sum = sum(obj['sum'] for obj in object_list)
            self.cnt = sum(obj['cnt'] for obj in object_list)
        else:
            self.reset()


class MediaType(str, Enum):
    IMAGE_PIL = "pillow"
    FIGURE_MPL = "matplotlib"
    FIGURE_PX = "plotly"
    TEXT = "str"


@dataclass
class MediaData:
    media_data: Any
    media_type: MediaType
    media_note: str


class IMediaProducer(nn.Module, Interface):

    def on_epoch_start(self):
        pass

    def forward(self, mb: IMiniBatch) -> Sequence[MediaData]:
        pass

    def on_epoch_end(self):
        pass


class MediaManager:

    def __init__(self, medias: Sequence[IMediaProducer]):
        self.producers = medias
        self.output_cache = []

    def get_latest_output(self) -> Sequence[MediaData]:
        return self.output_cache

    def update_and_sync(self, mb:IMiniBatch):
        if 0 == len(self.producers):
            return []

        medias = []

        for producer in self.producers:
            media = producer.forward(mb)
            if not isinstance(media, Sequence) and isinstance(media[0], MediaData):
                raise TypeError(f"Media Producer {producer.__class__.__name__} is expected to output a sequence of MediaData, got {type(media)}")
            medias += media

        if dist.is_initialized():
            object_list = [None for _ in range(dist.get_world_size())]
            dist.gather_object(
                obj=medias,
                object_gather_list=object_list,
                dst=0,
            )
            medias = []
            if dist.get_rank() == 0:
                for obj in object_list:
                    medias += obj

        self.output_cache = medias

    def notify(self, events):
        for producer in self.producers:
            getattr(producer, events)()


def _find_free_port():
    """ find a not used port for DDP.

    https://stackoverflow.com/a/45690594
    """
    import socket
    from contextlib import closing
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(('', 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def _subprocess(rank, loop: Union["EvalLoop", "TrainLoop"]):
    """spawned sub processes entrypoint function.

    a :class:`EvalLoop` or :class:`TrainLoop` will spawn multiple processes if
    the :class:`TopLevelModule` it attacted to are told to move its parameters
    to multiple devices by :meth:`TopLevelModule.to`. This implies a distributed
    execution of the loop. The sub processes will then init a communication
    group, actually place model to the target devices, wrap it with torch's
    :class:``DistributedDataParallel`` wrapper, and call the loop's ``__call__``
    again like a single process execution. Finally, it cleans up context and
    exit to the spawning point of the father process.

    Args:
        rank (int): the index of the subprocess, starting from ``0`` to ``len(distributed_devices) - 1``.
        loop (EvalLoop | TrainLoop): the entire loop context is passed to the subprocess.
    """

    model = loop.model
    devices = model.distributed_devices
    model.to(devices[rank])
    model.distributed_devices = []
    model.ddp_rank = rank
    model.ddp_enabled = True

    os.environ.setdefault("MASTER_ADDR", "127.0.0.1")
    os.environ.setdefault("MASTER_PORT", str(model.ddp_port))

    dist.init_process_group(
        backend=os.environ.get("T3W_DDP_BACKEND", 'nccl'),
        rank=rank,
        world_size=len(devices),
    )
    dist.barrier()

    if isinstance(loop, TrainLoop):
        model._fix_optim_states()
        loop.ddp_model = DistributedDataParallel(
            model,
            static_graph=os.environ.get("T3W_STATIC_GRAPH", '1')=='1',
            )

    loop()

    dist.destroy_process_group()


@dataclass
class _TrainingProgress:
    """a class that count the total training steps and epochs number of the model.

    This progress is a part of the state_dict of the :class:`TopLevelModule`,
    and :class:`TrainLoop` makes use of it to resume training,
    :class:`SaveBestModelsSideEffect` makes use of it to label the checkpoint
    saving filename. So it is "the training progress" instead of "a training
    progress" of the model.
    """

    step: int = 0
    """number of the total times the ``optim.step`` method has been called.

    Note:
        It is not about how many actual iteration the for loop has run, it is
        the step of the optimizer has updated the model. Consider a gradient
        accumulation case, only after multiple "iteration steps" will the
        optimizer step once, and the :meth:`inc_step` will also be called only
        once.
    """

    epoch: int = 0
    """number of the total training epoch."""

    steps_per_epoch: int = 0

    total_epochs: int = 0

    def inc_step(self):
        """increase the training step by 1."""
        self.step += 1

    def inc_epoch(self):
        """increase the training epoch by 1."""
        self.epoch += 1


class TopLevelModule(nn.Module):
    """A central API that manages the ``nn.Module`` related features.

    This top level module helps its owner loop on infrastructure the code, and helps user by providing useful low-level utilities, including
        * manages model checkpoint saving and loading,
        * moves user_model to other device(s) and trigger DDP execution mode,
        * computes losses and metrics specified by the loop.

    Note:
        We delegate losses computation to :class:`TrainLoop` (see
        :meth:`TrainLoop.__init__`), while the loop delegates it further to
        :meth:`TopLevelModule.forward`. This is clever because torch's
        ``DataDistributedParallel`` class wraps the top level module instead of
        the user's model therefore always has a loss tensor as output and be
        able to find unused parameters during training, while user can stick to
        the suggested standard OOP paradigm in the userspace by operating on the
        :class:`IMiniBatch` instance.
    """

    user_model: nn.Module
    """User's model consumes and updates the input :class:`IMiniBatch` data in its forward method, and optionally return a model specific regularizer loss. """

    optim:Optimizer
    """Stores user defined optimizer, the calling is further delegated to :class:`TrainLoop`. """

    lr_scheduler:_LRScheduler
    """Stores user defined learning rate scheduler, the calling is further delegated to :class:`TrainLoop`. """

    training_progress:_TrainingProgress
    """Stores training process of current model. It is part of the checkpoint state_dict. Modification in userspace is not intended."""

    regularizer_reweight:float
    """Stores the weight of regularizer loss if the model has one. It is **NOT** part of the checkpoint state_dict.  Modification in userspace at training should be done through the ``loop`` argument in event :meth:`on_train_step_started`."""


    def __init__(
            self,
            model:nn.Module,
            optim:Optimizer = None,
            lr_scheduler:_LRScheduler = None,
            regularizer_reweight: float = 1.
        ) -> None:
        """

        Args:
            model (nn.Module): user defined model.
            optim (Optimizer, optional): user defined optimizer. Defaults to None.
            lr_scheduler (_LRScheduler, optional): user defined lr_scheduler. Defaults to None.
            regularizer_reweight (float, optional): the weight of the regularizer_loss if any. Defaults to 1.
        """
        super().__init__()
        self.user_model:nn.Module = model
        self.optim:Optimizer = optim
        self.lr_scheduler:_LRScheduler = lr_scheduler
        self.training_progress = _TrainingProgress()
        self.regularizer_reweight:float = regularizer_reweight

        self._device = torch.device("cpu")
        self.distributed_devices = []
        self.ddp_enabled = False
        self.ddp_rank = 0
        self.ddp_port = 0

    def to(self, device):
        """move model to target device or trigger DDP execution if multiple target devices are provided.

        Warning:
            When multiple target devices are specified, the parameters will not be moved until the subprocesses
            are spawned by the :class:`TrainLoop` or :class:`EvalLoop`.

        Args:
            device (str): "cuda:0", "cuda:0,1", "cuda:0-3" are all valid inputs.
        """
        devices = self._parse_multi_device(device)
        if 1 == len(devices):
            self.user_model.to(device)
            self._device = torch.device(device)
        else:
            self.distributed_devices = devices
            self.ddp_port = _find_free_port()
        return self

    def forward(
            self,
            mb: IMiniBatch,
            losses: Mapping[str, ILoss] = None,
            metrics: Mapping[str, IMiniBatchMetric] = None,
            step_dict: StepReturnDict = None,
        ) -> Union[MiniBatchFloats, IMiniBatch]:
        """

        In training mode, this will call ``self.user_model(mb)``, compute ``losses`` and ``metrics``, collect results to fill in ``step_dict``, and return weighted sum of all losses (to be backward).

        In evaluation mode, this will call ``self.user_model(mb)`` and return mb.

        Args:
            mb (IMiniBatch): _description_
            losses (Mapping[str, ILoss], optional): _description_. Defaults to None.
            metrics (Mapping[str, IMiniBatchMetric], optional): _description_. Defaults to None.
            step_dict (StepReturnDict, optional): _description_. Defaults to None.

        Returns:
            Union[MiniBatchFloats, IMiniBatch]: _description_
        """
        mb.to(self.device)
        mb.model = self
        if self.training:
            train_loss:Tensor = 0.
            regularizer_loss = self.user_model(mb)
            if regularizer_loss is not None:
                train_loss += self.regularizer_reweight * regularizer_loss
                step_dict["losses"]["regularizer_loss"] = (self.regularizer_reweight, regularizer_loss)
            for loss_name, loss_obj in losses.items():
                step_dict["losses"][loss_name] = (loss_obj.loss_reweight, loss_obj(mb))
                train_loss += loss_obj.loss_reweight * step_dict["losses"][loss_name][1]
            for metric_name, metric_obj in metrics.items():
                step_dict["metrics"][metric_name] = metric_obj(mb)
            return train_loss
        else:
            self.user_model(mb)
            return mb

    @property
    def device(self):
        """current device (not supporting model parallelzation)"""
        try:
            return next(self.user_model.parameters()).device
        except StopIteration:
            return self._device

    def save(self, path):
        """save current training states to the disk.

        The states to be maintained include:
            * training progress
            * user model state_dict
            * user optimizer state_dict
            * user lr_scheduler state_dict
            * current random states of ``random``, ``np.random`` and ``torch.random``.

        Warning:
            No parent directory will be produced if not exist.

        Args:
            path (str): save path.
        """
        with gzip.open(path, 'wb') as ckpt_file:
            torch.save(
                {
                    "training_progress": asdict(self.training_progress),
                    "model": self.user_model.state_dict(),
                    "optim": self.optim.state_dict() if self.optim is not None else None,
                    "lr_scheduler": self.lr_scheduler.state_dict() if self.lr_scheduler is not None else None,
                    "random_states": {
                        "random": random.getstate(),
                        "np": np.random.get_state(),
                        "torch": torch.random.get_rng_state(),
                    }
                },
                ckpt_file
            )

    def load(self, path, strict=True):
        """load a previous top level model checkpoint

        Args:
            path (str): load file path
            strict (bool, optional): passed on to user_model's ``load_state_dict(strict=?)`` option. Defaults to True.

        Returns:
            _IncompitableKeys: returned incompitable keys of user_model's ``load_state_dict()`` method.
        """
        with gzip.open(path, 'rb') as ckpt_file:
            ckpt_dict = torch.load(ckpt_file, map_location=str(self.device))
        self.training_progress = _TrainingProgress(**ckpt_dict['training_progress'])
        model_load_return = self.user_model.load_state_dict(ckpt_dict['model'], strict=strict)
        if self.optim is not None:
            self.optim.load_state_dict(ckpt_dict['optim'])
        if self.lr_scheduler is not None:
            self.lr_scheduler.load_state_dict(ckpt_dict['lr_scheduler'])
        random.setstate(ckpt_dict['random_states']['random'])
        np.random.set_state(ckpt_dict['random_states']['np'])
        torch.random.set_rng_state(ckpt_dict['random_states']['torch'].cpu())
        return model_load_return

    def _fix_optim_states(self):
        """
        _LRScheduler of PyTorch makes a monkey patch on the optimizer to help detect the proper order of calling
        ``optim.step`` and ``lr_scheduler.step``, this patch will be lost though in a multi-processing spawning.
        We reproduce this monkey patch in the sub process to repress the warning information.
        """
        def with_counter(method):
            if getattr(method, '_with_counter', False):
                # `optimizer.step()` has already been replaced, return.
                return method

            # Keep a weak reference to the optimizer instance to prevent
            # cyclic references.
            instance_ref = weakref.ref(method.__self__)
            # Get the unbound method for the same purpose.
            func = method.__func__
            cls = instance_ref().__class__
            del method

            @wraps(func)
            def wrapper(*args, **kwargs):
                instance = instance_ref()
                instance._step_count += 1
                wrapped = func.__get__(instance, cls)
                return wrapped(*args, **kwargs)

            # Note that the returned function here is no longer a bound method,
            # so attributes like `__func__` and `__self__` no longer exist.
            wrapper._with_counter = True
            return wrapper
        self.optim.step = with_counter(self.optim.step)
        self.optim._step_count = 0

    @staticmethod
    def _parse_multi_device(devices:str) -> List[torch.device]:
        """
        Args:
            - devices: e.g. "cuda", "cuda:0", "cuda:0,1", "cuda:0-2", "cuda:0-1,3".
        Return:
            List["torch.device"]
        """

        if isinstance(devices, (int, torch.device)): return ([devices], None)

        # determine device type
        if devices[:3] == "cpu": device_type = "cpu"
        elif devices[:4] == "cuda": device_type = "cuda"
        else: raise ValueError(f"Unsupported devices value: {repr(devices)}")

        # determine device indices
        idid = devices.find(":")
        if -1 == idid: out = [torch.device(device_type, 0)]
        else:
            out = []
            for indices in devices[idid+1:].split(","):
                if -1 != indices.find("-"):
                    s, e = indices.split("-")
                    for i in range(int(s), int(e) + 1):
                        out.append(torch.device(device_type, i))
                else:
                    out.append(torch.device(device_type, int(indices)))
            if 0 == len(out): raise ValueError("Empty devices indices.")
        return out


class ProbeData(nn.Module):

    def __init__(self, breakpoint:bool = True) -> None:
        super().__init__()
        self.breakpoint = breakpoint

    def forward(self, data: "IMiniBatch"):
        if self.breakpoint:
            pprint(data, max_depth=2, max_length=16)
            breakpoint()
            exit()


class EvalLoop:

    dataset: IDataset
    model: TopLevelModule
    batch_size: int
    loader: DataLoader
    metrics: Mapping[str, IDatasetMetric]

    def __init__(self,
                 dataset: IDataset,
                 model: TopLevelModule = None,
                 batch_size: int = None,
                 metrics: Mapping[str, IDatasetMetric] = dict(),
                 medias: Sequence[IMediaProducer] = [],
                 side_effects: Sequence["ISideEffect"] = [],
                 breakpoint: bool = False,
                 ) -> None:
        self.dataset = dataset
        self.model = model or TopLevelModule(ProbeData(breakpoint=breakpoint))
        self.metrics = metrics
        self.medias = MediaManager(medias)
        self.handle:_EventHandler = _EventHandler(side_effects)
        self.loader:DataLoader = None
        self.batch_size:int = batch_size or self.dataset.datum_type.val_batch_size

    def __call__(self) -> None:

        if self.model.distributed_devices:
            return mp.start_processes(_subprocess, (self,), nprocs=len(self.model.distributed_devices), join=True, start_method='spawn')

        self.loader = DataLoader(
            self.dataset,
            sampler=DistributedSampler(self.dataset, shuffle=False) if self.model.ddp_enabled else None,
            batch_size=self.batch_size,
            collate_fn=self.dataset.datum_type.collate,
            num_workers=self.dataset.datum_type.num_workers,
            pin_memory=self.model.device.type!="cpu",
            pin_memory_device=str(self.model.device) if self.model.device.type!="cpu" else "",
        )

        self.model.train(False)
        for metric in self.metrics.values():
            metric.reset()
        self.medias.notify("on_epoch_start")
        self.handle('on_eval_started', self)
        for step, mb in enumerate(self.loader):
            self.handle('on_eval_step_started', self, step, mb)
            self.model(mb)
            self.mark_padding(step, mb)
            for metric in self.metrics.values():
                metric.update(mb)
            self.medias.update_and_sync(mb)
            self.handle('on_eval_step_finished', self, step, mb)
        for metric in self.metrics.values():
            metric.synchronize()
        self.medias.notify("on_epoch_end")
        self.handle('on_eval_finished', self)
        self.model.train(True)

    @property
    def metric_values(self) -> Dict[str, float]:
        return {k:v.eval() for k, v in self.metrics.items()}

    def mark_padding(self, step, mb: IMiniBatch):
        if dist.is_initialized():
            processed = self.batch_size * (step * dist.get_world_size() + dist.get_rank())
        else:
            processed = self.batch_size * step
        mb._nopad_batch_size = max(min(len(self.dataset) - processed, self.batch_size), 0)


def _seed_worker(worker_id):
    worker_seed = torch.initial_seed() % 2**32
    np.random.seed(worker_seed)
    random.seed(worker_seed)


class TrainLoop:

    dataset: IDataset
    model: TopLevelModule
    batch_size: int
    num_acc_grad: int
    epochs: int
    loader: DataLoader
    losses: Mapping[str, ILoss]
    metrics: Mapping[str, IDatasetMetric]
    eval_loop: EvalLoop

    def __init__(
            self,
            dataset: IDataset,
            model: TopLevelModule,
            losses: Mapping[str, ILoss] = dict(),
            metrics: Mapping[str, IMiniBatchMetric] = dict(),
            medias: Sequence[IMediaProducer] = [],
            batch_size: Optional[int] = None,
            sampler: Sampler = None,
            num_acc_grad: int = 1,
            epochs: int = 100,
            iter_per_epoch: Optional[int] = None,
            epoch_per_eval: int = 1,
            eval_loop: Optional[EvalLoop] = None,
            side_effects: Sequence["ISideEffect"] = [],
        ) -> None:
        """

        Args:
            model (TopLevelModule): the top level model.
            dataset (IDataset): train split of the dataset.
            losses (Mapping[str, ILoss]): losses to be evaluated.
            metrics (Mapping[str, IMiniBatchMetric]): metrics to be evaluated
            batch_size (Optional[int], optional): train batch size. Defaults to the ``train_batch_size`` static attribute of :class:`IDatum`.
            sampler (Sampler, optional): a custom sampler instance for the dataloader. Defaults to ``RandomSampler`` or ``DistributedSampler``.
            num_acc_grad (int, optional): step interval to apply gradient descent. Defaults to 1.
            epochs (int, optional): total training epochs. Defaults to 100.
            iter_per_epoch (Optional[int], optional): manually define iteration number of an epoch. Defaults to ``len(dataloader)``.
            epoch_per_eval (int, optional): epoch interval to apply ``eval_loop``. Defaults to 1.
            eval_loop (Optional[EvalLoop], optional): an :class:`EvalLoop` instance. Defaults to None.
            side_effects (Sequence[ISideEffect], optional): Defaults to [].
        """
        self.dataset = dataset
        self.model:TopLevelModule = model
        self.losses = losses
        self.metrics = metrics
        self.medias = MediaManager(medias)
        self.batch_size:int = batch_size or self.dataset.datum_type.train_batch_size
        """The normal mini batch size during training. """
        self.num_acc_grad = num_acc_grad
        self.iter_per_epoch = iter_per_epoch
        self.epoch_per_eval = epoch_per_eval
        self.eval_loop = eval_loop
        self.handle:_EventHandler = _EventHandler(side_effects)
        self.sampler = sampler
        self.model_forward = model.forward
        self.loader:DataLoader = None
        self.ddp_model: DistributedDataParallel = None
        self.epochs = epochs

    def __call__(self):
        """Implement iter_based and epoch_based train_loop with hooks."""


        if self.model.distributed_devices:
            return mp.start_processes(_subprocess, (self,), nprocs=len(self.model.distributed_devices), join=True, start_method='spawn')


        for loss in self.losses.values(): loss.to(self.model.device)
        for metric in self.metrics.values(): metric.to(self.model.device)


        if self.model.ddp_enabled:
            self.model_forward = self.ddp_model.forward

            loader_kwargs = dict(
                sampler=self.sampler or DistributedSampler(self.dataset, shuffle=True, drop_last=True),
            )
        else:
            g = torch.Generator()
            g.manual_seed(torch.initial_seed())
            loader_kwargs = dict(
                shuffle=self.sampler is None,
                sampler=self.sampler,
                generator=g,
                drop_last=True,
                worker_init_fn=_seed_worker,
            )

        if isinstance(loader_kwargs['sampler'], DistributedSampler):
            set_epoch = lambda epoch: loader_kwargs['sampler'].set_epoch(epoch)
        else:
            set_epoch = lambda _: None

        self.loader = DataLoader(
            self.dataset,
            batch_size=self.batch_size,
            num_workers=self.dataset.datum_type.num_workers,
            collate_fn=self.dataset.datum_type.collate,
            pin_memory=self.model.device.type!="cpu",
            pin_memory_device=str(self.model.device) if self.model.device.type!="cpu" else "",
            **loader_kwargs
        )

        start_epoch = self.model.training_progress.epoch
        iter_loader = iter(())
        data_size = self.iter_per_epoch or len(self.loader)
        self.progress.steps_per_epoch = data_size // self.num_acc_grad
        self.handle('on_train_started', self)
        for epoch in range(start_epoch, self.epochs):
            self.handle('on_train_epoch_started', self, epoch)
            self.medias.notify("on_epoch_start")
            for step in range(data_size):
                try: mb = next(iter_loader)
                except StopIteration:
                    set_epoch(self.model.training_progress.step // len(self.loader))
                    iter_loader = iter(self.loader)
                    mb = next(iter_loader)
                self.handle('on_train_step_started', self, step, mb)
                step_return = self.step(mb)
                self.handle('on_train_step_finished', self, step, mb, step_return)
            if self.model.lr_scheduler is not None and len(step_return["losses"]):
                self.model.lr_scheduler.step()
            self.model.training_progress.inc_epoch()
            if self.eval_loop and epoch % self.epoch_per_eval == 0: self.eval_loop()
            self.medias.notify("on_epoch_end")
            self.handle('on_train_epoch_finished', self, epoch)

    def step(self, mb: IMiniBatch):
        """Implement a single train step, with gradiant accumulate."""
        step_dict = dict(losses=dict(), metrics=dict())
        training_progress = self.model.training_progress
        if training_progress.step % self.num_acc_grad == 0:
            self.model.optim.zero_grad()
        train_loss = self.model_forward(mb, self.losses, self.metrics, step_dict)
        if isinstance(train_loss, Tensor):
            (train_loss / self.num_acc_grad).backward()
            training_progress.inc_step()
            if training_progress.step % self.num_acc_grad == 0:
                self.model.optim.step()
        self.medias.update_and_sync(mb)
        return step_dict

    @property
    def progress(self):
        return self.model.training_progress

    @property
    def epochs(self) -> int:
        return self.progress.total_epochs

    @epochs.setter
    def epochs(self, value:int):
        self.progress.total_epochs = value
        return value


class _EventHandler:

    def __init__(self, side_effects: Sequence["ISideEffect"] = []) -> None:
        self.side_effects = side_effects

    def __call__(self, event: str, *args, **kwargs):
        for side_effect in self.side_effects:
            if dist.is_initialized():
                abort = 0
                if dist.get_rank() == 0 or side_effect.is_distributed:
                    try: getattr(side_effect, event)(*args, **kwargs)
                    except:
                        abort = 1
                object_list = [None for _ in range(dist.get_world_size())]
                dist.all_gather_object(object_list, abort)
                if sum(object_list):
                    exit()
            else:
                getattr(side_effect, event)(*args, **kwargs)


class ISideEffect(Interface):

    is_distributed:bool = False
    """if False, the event will only be invoked on the rank 0 sub-process; otherwise on all sub-processes."""

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__ = d

    def on_eval_started(self, loop: "EvalLoop"):
        pass

    def on_eval_step_started(self, loop: "EvalLoop", step: int, mb: "IMiniBatch"):
        pass

    def on_eval_step_finished(self, loop: "EvalLoop", step: int, mb:"IMiniBatch"):
        pass

    def on_eval_finished(self, loop: "EvalLoop"):
        pass

    def on_train_started(self, loop: "TrainLoop"):
        pass

    def on_train_epoch_started(self, loop: "TrainLoop", epoch: int):
        pass

    def on_train_step_started(self, loop: "TrainLoop", step: int, mb: "IMiniBatch"):
        pass

    def on_train_step_finished(self, loop: "TrainLoop", step: int, mb: "IMiniBatch", step_return: StepReturnDict):
        pass

    def on_train_epoch_finished(self, loop: "TrainLoop", epoch: int):
        pass

    def on_train_finished(self, loop: "TrainLoop"):
        pass


def command(
    model:Type[nn.Module] = ProbeData,
    optim:Type[Optimizer] = None,
    lr_scheduler:Type[_LRScheduler] = None,
):
    def decorator(main):
        @wraps(main)
        def wrapper(*args, **kw):

            return main(*args, **kw)
        return wrapper
    return decorator



import os
import uuid
from contextlib import contextmanager
from datetime import datetime


def generate_run_hash(hash_length=24) -> str:
    full_hash = datetime.now().strftime("%y%m%d%H%M%S") + uuid.uuid4().hex
    return full_hash[:hash_length]


@contextmanager
def work_directory(path: str):
    curr = os.getcwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(curr)



from typing import Callable
import sys

_runtime_subcommand_registry = dict()


def cmd(subcommand: Callable):

    name = subcommand.__name__

    if name in _runtime_subcommand_registry:
        raise RuntimeError(f"subcommand {name} duplicated.")

    _runtime_subcommand_registry[name] = subcommand

    return subcommand


def main():

    if len(sys.argv) > 1 and sys.argv[1] in _runtime_subcommand_registry:
        _runtime_subcommand_registry[sys.argv[1]](sys.argv[2:])
    else:
        print(f"Usage: python {sys.argv[0]} {'|'.join(list(_runtime_subcommand_registry.keys()))}")



""" This code is modified from https://github.com/nmhkahn/torchsummaryX, credit to the orignal author Namhyuk Ahn.
"""

from collections import OrderedDict
import numpy as np
import pandas as pd
import torch

# Some modules do the computation themselves using parameters or the parameters of children, treat these as layers
layer_modules = (torch.nn.MultiheadAttention, )

def summary(model, x, *args, layer_modules=layer_modules, print_summary=True, **kwargs):
    """Summarize the given input model.
    Summarized information are 1) output shape, 2) kernel shape,
    3) number of the parameters and 4) operations (Mult-Adds)
    Args:
        model (Module): Model to summarize
        x (Tensor): Input tensor of the model with [N, C, H, W] shape
                    dtype and device have to match to the model
        layer_modules: Some modules do the computation themselves using
                    parameters or the parameters of children, treat these as layers.
        args, kwargs: Other argument used in `model.forward` function
    """
    def register_hook(module):
        def hook(module, inputs, outputs):
            cls_name = str(module.__class__).split(".")[-1].split("'")[0]
            module_idx = len(summary)

            # Lookup name in a dict that includes parents
            module_name = str(module_idx)
            for name, item in module_names.items():
                if item == module:
                    module_name = name
                    break
            key = "{}_{}".format(module_idx, name)

            info = OrderedDict()
            info["id"] = id(module)
            if isinstance(outputs, (list, tuple)):
                try:
                    info["out"] = list(outputs[0].size())
                except AttributeError:
                    # pack_padded_seq and pad_packed_seq store feature into data attribute
                    info["out"] = list(outputs[0].data.size())
            else:
                info["out"] = list(outputs.size())

            info["ksize"] = "-"
            info["inner"] = OrderedDict()
            info["params_nt"], info["params"], info["macs"] = 0, 0, 0
            for name, param in module.named_parameters():
                info["params"] += param.nelement() * param.requires_grad
                info["params_nt"] += param.nelement() * (not param.requires_grad)

                if name == "weight":
                    ksize = list(param.size())
                    # to make [in_shape, out_shape, ksize, ksize]
                    if len(ksize) > 1:
                        ksize[0], ksize[1] = ksize[1], ksize[0]
                    info["ksize"] = ksize

                    # ignore N, C when calculate Mult-Adds in ConvNd
                    if "Conv" in cls_name:
                        info["macs"] += int(param.nelement() * np.prod(info["out"][2:]))
                    else:
                        info["macs"] += param.nelement()

                # RNN modules have inner weights such as weight_ih_l0
                elif "weight" in name:
                    info["inner"][name] = list(param.size())
                    info["macs"] += param.nelement()

            # if the current module is already-used, mark as "(recursive)"
            # check if this module has params
            if list(module.named_parameters()):
                for v in summary.values():
                    if info["id"] == v["id"]:
                        info["params"] = "(recursive)"

            if info["params"] == 0:
                info["params"], info["macs"] = "-", "-"

            summary[key] = info

        # ignore Sequential and ModuleList and other containers
        if isinstance(module, layer_modules) or not module._modules:
            hooks.append(module.register_forward_hook(hook))

    module_names = get_names_dict(model)

    hooks = []
    summary = OrderedDict()

    model.apply(register_hook)
    try:
        with torch.no_grad():
            model(x) if not (kwargs or args) else model(x, *args, **kwargs)
    except Exception:
        # This can be usefull for debugging
        print("Failed to run torchsummaryX.summary, printing sizes of executed layers:")
        df = pd.DataFrame(summary).T
        print(df)
        raise
    finally:
        for hook in hooks:
            hook.remove()

    # Use pandas to align the columns
    pd.set_option('display.max_rows', 500)
    pd.set_option('display.max_columns', 500)
    pd.set_option('display.width', 1000)
    df = pd.DataFrame(summary).T

    df["Mult-Adds"] = pd.to_numeric(df["macs"], errors="coerce")
    df["Params"] = pd.to_numeric(df["params"], errors="coerce")
    df["Non-trainable params"] = pd.to_numeric(df["params_nt"], errors="coerce")
    df = df.rename(columns=dict(
        ksize="Kernel Shape",
        out="Output Shape",
    ))
    df_sum = df[["params_nt", "Mult-Adds", "Params", "Non-trainable params"]].sum() # https://github.com/nmhkahn/torchsummaryX/pull/21
    df.index.name = "Layer"

    df = df[["Kernel Shape", "Output Shape", "Params", "Mult-Adds"]]
    max_repr_width = max([len(row) for row in df.to_string().split("\n")])

    df_total = pd.DataFrame(
        {"Total params": (df_sum["Params"] + df_sum["params_nt"]),
        "Trainable params": df_sum["Params"],
        "Non-trainable params": df_sum["params_nt"],
        "Mult-Adds": df_sum["Mult-Adds"]
        },
        index=['Totals']
    ).T

    if print_summary:
        option = pd.option_context(
            "display.max_rows", 600,
            "display.max_columns", 10,
            "display.float_format", pd.io.formats.format.EngFormatter(use_eng_prefix=True)
        )
        with option:
            print("="*max_repr_width)
            print(df.replace(np.nan, "-"))
            print("-"*max_repr_width)
            print(df_total)
            print("="*max_repr_width)

    return df, df_total

def get_names_dict(model):
    """Recursive walk to get names including path."""
    names = {}

    def _get_names(module, parent_name=""):
        for key, m in module.named_children():
            cls_name = str(m.__class__).split(".")[-1].split("'")[0]
            num_named_children = len(list(m.named_children()))
            if num_named_children > 0:
                name = parent_name + "." + key if parent_name else key
            else:
                name = parent_name + "." + cls_name + "_"+ key if parent_name else key
            names[name] = m

            if isinstance(m, torch.nn.Module):
                _get_names(m, parent_name=name)

    _get_names(model)
    return names



from pathlib import Path
from t3w.core import *
from t3w.core import EvalLoop
from t3w.utils.verbose import millify
from functools import cached_property


class SaveBestModelsSideEffect(ISideEffect):

    def __init__(
            self,
            metric_name: str,
            num_max_keep: int,
            save_path_prefix: str="./",
        ) -> None:
        super().__init__()
        self.metric_name = metric_name
        self.num_max_keep = num_max_keep
        self.save_path_prefix = save_path_prefix
        self.history:List[Dict[Literal["metric_value", "saved_path"], Any]] = []
        Path(save_path_prefix + "-").parent.mkdir(parents=True, exist_ok=True)

    def on_train_started(self, loop: TrainLoop):
        assert self.metric_name in loop.eval_loop.metrics.keys(), "metric_name for saving model does not exist in eval_loop"
        self._guard_duplicated_saving_path

    def on_eval_started(self, loop: EvalLoop):
        self._guard_duplicated_saving_path

    @cached_property
    def _guard_duplicated_saving_path(self):
        matched_existing_paths = [path for path in glob(self.save_path_prefix+"*") if self.metric_name in path and ".pt" in path]

        if len(matched_existing_paths):
            raise FileExistsError(f"{self.__class__.__name__}: found existing checkpoint files matching {self.save_path_prefix}*")

        return True

    def on_eval_finished(self, loop: EvalLoop):

        if not (self.num_max_keep > 0): return

        metric_value = loop.metric_values[self.metric_name]
        higher_better = loop.metrics[self.metric_name].minibatch_metric.higher_better
        better = lambda a, b: (higher_better and a >= b) or (not higher_better and a <= b)

        if self.history:
            sorted_history = sorted(
                self.history,
                key=lambda item: item["metric_value"],
                reverse=higher_better
            )
            topk = sorted_history[:self.num_max_keep][-1]
            if not better(metric_value, topk["metric_value"]):
                return
            if len(sorted_history) >= self.num_max_keep:
                for item in sorted_history[self.num_max_keep-1:]:
                    os.remove(item['saved_path'])
                self.history = sorted_history[:self.num_max_keep-1]

        short_bar = "-" if Path(self.save_path_prefix + "-").name != "-" else ""
        current = dict(
            metric_value=metric_value,
            saved_path=f"{self.save_path_prefix}{short_bar}ep{loop.model.training_progress.epoch}-step{millify(loop.model.training_progress.step)}-{self.metric_name}={metric_value:.5f}.pt.gz".replace("--ep", "-ep"),
        )
        loop.model.save(current["saved_path"])
        self.history.append(current)



from t3w.core import *


class TqdmSideEffect(ISideEffect):

    def __init__(self, postfix_keys:Sequence[str]= [], smoothing:float = 0.9) -> None:
        """show a tqdm progress bar for training and evaluation, write eval metrics at the end of eval epoch.

        Args:
            postfix_keys (Sequence[str], optional): control the losses and metrics to be displayed on the bar. Empty sequence will show all.
            smoothing (float, optional): low pass filter value. Defaults to 0.9.
        """
        super().__init__()
        from tqdm import tqdm
        self.tqdm = tqdm
        self.postfix_keys = postfix_keys
        self.tqdm_kwargs = dict(smoothing=smoothing, leave=False, dynamic_ncols=True)

    def on_eval_started(self, loop: "EvalLoop"):
        self.eval_pbar = self.tqdm(desc=f"Eval {loop.model.training_progress.epoch}", total=len(loop.loader), **self.tqdm_kwargs)

    def on_eval_step_finished(self, loop: "EvalLoop", step: int, mb: "IMiniBatch"):
        self.eval_pbar.update()

    def on_eval_finished(self, loop: "EvalLoop"):
        self.eval_pbar.close()
        metrics_dict = dict()
        metrics_dict["epoch"] = loop.model.training_progress.epoch
        metrics_dict.update(loop.metric_values)
        self.tqdm.write(repr(metrics_dict))

    def on_train_started(self, loop: "TrainLoop"):
        self.train_epoch_pbar = self.tqdm(desc="Epochs", total=loop.epochs, **self.tqdm_kwargs)

    def on_train_epoch_started(self, loop: "TrainLoop", epoch: int):
        self.train_step_pbar = self.tqdm(desc=f"Train {epoch}", total=(loop.iter_per_epoch or len(loop.loader)), leave=False, dynamic_ncols=True)

    def on_train_step_finished(self, loop: "TrainLoop", step: int, mb: "IMiniBatch", step_return: StepReturnDict):
        self.train_step_pbar.update()

        step_postfix = dict()
        postfix_keys = self.postfix_keys or (list(step_return["losses"].keys()) + list(step_return["metrics"].keys()))
        for key in postfix_keys:
            if key in step_return["losses"]:
                loss_reweight, loss_raw_value = step_return["losses"][key]
                loss_raw_value = f"{loss_raw_value:.4f}"[:6]
                step_postfix[key] = f"{loss_reweight}*{loss_raw_value}"
            elif key in step_return["metrics"]:
                metric_value = f"{step_return['metrics'][key]:.4f}"[:6]
                step_postfix[key] = f"{metric_value}"
            else:
                pass  # missing key ignored
        self.train_step_pbar.set_postfix(step_postfix)

    def on_train_epoch_finished(self, loop: "TrainLoop", epoch: int):
        self.train_step_pbar.close()
        self.train_epoch_pbar.update()

    def on_train_finished(self, loop: "TrainLoop"):
        self.train_epoch_pbar.close()



import os
from setuptools import setup, find_packages


__version__ = '0.3-alpha1'

with open("README.md", 'r') as readme_file:
    long_description = readme_file.read()

setup(
    name="t3w",
    version=__version__,
    author="Yuyao Huang",
    author_email="huangyuyao@outlook.com",
    url="https://github.com/tjyuyao/t3w",
    description="Typed Thin PyTorch Wrapper",
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    install_requires=[
        "jaxtyping",
        "pandas",
        "aim",
        "tqdm",
        "rich",
        "docstring_parser",
        "typing_inspect",
        "mgzip",
        # "typer[all]",
    ],
    extras_require={
        # "all": [
        #     "aim",
        #     "tqdm",
        # ]
    },
)



from t3w.core import *
from t3w.core import EvalLoop, IMiniBatch
from .patch import Run
from PIL import Image as PIL_Image
import aim
import matplotlib.figure
from functools import cached_property


class AimSideEffect(ISideEffect):

    def __init__(
            self,
            experiment: str,
            hparams_dict: Dict[str, Any],
            run_hash: str = None,
            description: str = None,
            repo: Union[str, Path] = "./",
            track_weights_every_n_steps: int = None,
        ) -> None:
        super().__init__()
        self.run_kwargs = dict(run_hash=run_hash, repo=repo, experiment=experiment)
        self.description = description
        self.hparams_dict = hparams_dict
        self.track_weights_every_n_steps = track_weights_every_n_steps

    @cached_property
    def run(self):
        from aim.storage.treeutils_non_native import convert_to_native_object

        run = Run(**self.run_kwargs)
        if self.description:
            run.description = self.description
        run['hparams'] = {k:convert_to_native_object(v, strict=False)
                          for k, v in self.hparams_dict.items()}

        return run

    def on_eval_finished(self, loop: EvalLoop):
        prog = loop.model.training_progress
        for name, value in loop.metric_values.items():
            self.run.track(value, name=f"eval/{name}", step=prog.step, epoch=prog.epoch)

    def on_train_step_finished(self, loop: TrainLoop, step: int, mb: "IMiniBatch", step_return: StepReturnDict):
        prog = loop.model.training_progress

        for name, (loss_reweight, loss_raw_value) in step_return['losses'].items():
            self.run.track(loss_raw_value, name=f"train/loss/{name}/raw", step=prog.step, epoch=prog.epoch)
            self.run.track(loss_reweight * loss_raw_value, name=f"train/loss/{name}/reweighted", step=prog.step, epoch=prog.epoch)

        for name, value in step_return['metrics'].items():
            self.run.track(value, name=f"train/metric/{name}", step=prog.step, epoch=prog.epoch)
        if self.track_weights_every_n_steps and (step + 1) % self.track_weights_every_n_steps == 0:
            from aim.sdk.adapters.pytorch import track_params_dists, track_gradients_dists
            track_params_dists(loop.model, self.run)
            track_gradients_dists(loop.model, self.run)

        self.handle_medias(loop.medias.get_latest_output(), step=prog.epoch, training=True)

    def on_eval_step_finished(self, loop: EvalLoop, step: int, mb: IMiniBatch):
        prog = loop.model.training_progress
        self.handle_medias(loop.medias.get_latest_output(), step=prog.epoch, training=False)

    def handle_medias(self, media_cache:Sequence[MediaData], step:NotImplemented, training:bool):
        media: MediaData
        for media in media_cache:
            if media.media_type in [MediaType.FIGURE_PX]:
                aim_figure = aim.Figure(media.media_data)
                self.run.track(aim_figure, name=media.media_type, step=step, context={"note": media.media_note, "training": training})
            elif media.media_type in [MediaType.FIGURE_MPL]:
                aim_image = aim.Image(fig2img(media.media_data))
                self.run.track(aim_image, name=media.media_type, step=step, context={"note": media.media_note, "training": training})
            else:
                raise NotImplementedError(f"Unsupported media type '{media.media_type}'.")


def fig2img(fig:matplotlib.figure.Figure) -> PIL_Image:
    """Convert a Matplotlib figure to a PIL Image and return it"""
    import io
    buf = io.BytesIO()
    fig.savefig(buf)
    buf.seek(0)
    img = PIL_Image.open(buf)
    return img



"""t3w.utils.argparse submodule is inspired and modified from the project https://pypi.org/project/typed-argument-parser/, credit to the original author. """
from __future__ import annotations

import argparse
import sys
from argparse import ArgumentTypeError
from collections.abc import Sequence
from dataclasses import dataclass, field
from functools import cached_property, partial
from inspect import Parameter, signature
from typing import (
    Any,
    Callable,
    Dict,
    get_type_hints,
    List,
    Literal,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from docstring_parser import Docstring, parse
from typing_inspect import (
    get_args as typing_inspect_get_args,
    get_origin as typing_inspect_get_origin,
    is_literal_type,
)

PRIMITIVES = (str, int, float, bool)


def get_argument_name(*name_or_flags) -> str:
    """Gets the name of the argument.

    :param name_or_flags: Either a name or a list of option strings, e.g. foo or -f, --foo.
    :return: The name of the argument (extracted from name_or_flags).
    """
    if "-h" in name_or_flags or "--help" in name_or_flags:
        return "help"

    if len(name_or_flags) > 1:
        name_or_flags = [n_or_f for n_or_f in name_or_flags if n_or_f.startswith("--")]

    if len(name_or_flags) != 1:
        raise ValueError(
            f"There should only be a single canonical name for argument {name_or_flags}!"
        )

    return name_or_flags[0].lstrip("-")


def get_literals(
    literal: Literal, variable: str
) -> Tuple[Callable[[str], Any], List[str]]:
    """Extracts the values from a Literal type and ensures that the values are all primitive types."""
    literals = list(get_args(literal))

    if not all(isinstance(literal, PRIMITIVES) for literal in literals):
        raise ArgumentTypeError(
            f'The type for variable "{variable}" contains a literal'
            f"of a non-primitive type e.g. (str, int, float, bool).\n"
            f"Currently only primitive-typed literals are supported."
        )

    str_to_literal = {str(literal): literal for literal in literals}

    if len(literals) != len(str_to_literal):
        raise ArgumentTypeError("All literals must have unique string representations")

    def var_type(arg: str) -> Any:
        if arg not in str_to_literal:
            raise ArgumentTypeError(
                f'Value for variable "{variable}" must be one of {literals}.'
            )

        return str_to_literal[arg]

    return var_type, literals


# TODO: remove this once typing_inspect.get_origin is fixed for Python 3.8, 3.9, and 3.10
# https://github.com/ilevkivskyi/typing_inspect/issues/64
# https://github.com/ilevkivskyi/typing_inspect/issues/65
def get_origin(tp: Any) -> Any:
    """Same as typing_inspect.get_origin but fixes unparameterized generic types like Set."""
    origin = typing_inspect_get_origin(tp)

    if origin is None:
        origin = tp

    if sys.version_info >= (3, 10) and isinstance(origin, UnionType):
        origin = UnionType

    return origin


# TODO: remove this once typing_insepct.get_args is fixed for Python 3.10 union types
def get_args(tp: Any) -> Tuple[type, ...]:
    """Same as typing_inspect.get_args but fixes Python 3.10 union types."""
    if sys.version_info >= (3, 10) and isinstance(tp, UnionType):
        return tp.__args__

    return typing_inspect_get_args(tp)


def boolean_type(flag_value: str) -> bool:
    """Convert a string to a boolean if it is a prefix of 'True' or 'False' (case insensitive) or is '1' or '0'."""
    if "true".startswith(flag_value.lower()) or flag_value == "1":
        return True
    if "false".startswith(flag_value.lower()) or flag_value == "0":
        return False
    raise ArgumentTypeError(
        'Value has to be a prefix of "True" or "False" (case insensitive) or "1" or "0".'
    )


class TupleTypeEnforcer:
    """The type argument to argparse for checking and applying types to Tuples."""

    def __init__(self, types: List[type], loop: bool = False):
        self.types = [boolean_type if t == bool else t for t in types]
        self.loop = loop
        self.index = 0

    def __call__(self, arg: str) -> Any:
        arg = self.types[self.index](arg)
        self.index += 1

        if self.loop:
            self.index %= len(self.types)

        return arg


def type_to_str(type_annotation: Union[type, Any]) -> str:
    """Gets a string representation of the provided type.

    :param type_annotation: A type annotation, which is either a built-in type or a typing type.
    :return: A string representation of the type annotation.
    """
    # Built-in type
    if type(type_annotation) == type:
        return type_annotation.__name__

    # Typing type
    return str(type_annotation).replace("typing.", "")


if sys.version_info >= (3, 10):
    from types import UnionType


# Constants
EMPTY_TYPE = get_args(List)[0] if len(get_args(List)) > 0 else tuple()
BOXED_COLLECTION_TYPES = {List, list, Set, set, Tuple, tuple}
UNION_TYPES = {Union} | ({UnionType} if sys.version_info >= (3, 10) else set())
OPTIONAL_TYPES = {Optional} | UNION_TYPES
BOXED_TYPES = BOXED_COLLECTION_TYPES | OPTIONAL_TYPES

InputType = TypeVar("InputType")
OutputType = TypeVar("OutputType")
FunctionType = Callable[[InputType], OutputType]
ClassType = Type[OutputType]


class ArgumentParser(argparse.ArgumentParser):
    @cached_property
    def _typed_group_args(self) -> Dict[str, OutputBuilder | OutputType]:
        return dict()

    def add_options_for(
        self,
        class_or_function: Union[FunctionType, ClassType],
        dest: str = None,
        partial: bool = False,
        include_args: Optional[List[str]] = None,
        ignore_args: Optional[List[str]] = [],
        help: str = None,
        prefix: bool = False,
        **defaults,
    ):
        """adds a group of arguments based on the parsing of the signature of ``class_or_function``.

        The group of arguments can be composed into an instance of the class_or_function type / output type

        Args:
            class_or_function: The class or function to run with the provided arguments.
            dest: The attribute name used in the result namespace if not None. Defaults to None.
            partial: The result attribute is a functools.partial function to be updated. This is forced True when mandatory arguments are listed in the ``ignore_args``. Defaults to False.
            include_args: A list of argument names to be parsed from the command line. Defaults to "All".
            ignore_args: A list of argument names not to be parsed from the command line. Defaults to [].
            **defaults: Overwrites the defaults in function signature.
        """

        # Infer prefix
        if dest and (prefix is True):
            prefix = dest + "_"
        elif isinstance(prefix, str):
            prefix = prefix + "_"
        else:
            prefix = ""

        # Preprocess defaults
        if 1 == len(defaults) and "defaults" in defaults:
            defaults = defaults["defaults"]
        defaults = {(prefix + name): value for name, value in defaults.items()}

        # Create an OutputBuilder to store parsed arguments
        if dest:
            builder = OutputBuilder(class_or_function, partial)
            self._typed_group_args[dest] = builder

        # Get the group interface to add arguments
        group = self.add_argument_group(dest, help)

        # Get signature from class or function
        sig = signature(class_or_function)

        # Parse class or function docstring in one line
        if (
            isinstance(class_or_function, type)
            and class_or_function.__init__.__doc__ is not None
        ):
            doc = class_or_function.__init__.__doc__
        else:
            doc = class_or_function.__doc__

        # Parse docstring
        docstring = parse(doc)

        # Get the description of each argument in the class init or function
        param_to_description = {
            param.arg_name: param.description for param in docstring.params
        }

        is_ignored = lambda param_name: (param_name in ignore_args) or (
            isinstance(include_args, Sequence) and param_name not in include_args
        )

        # Add arguments for command line processing
        for param_name, param in sig.parameters.items():
            # Skip **kwargs
            if dest and param.kind == Parameter.VAR_KEYWORD:
                builder.pass_extra_kwargs = True
                continue

            ignored_flag = is_ignored(param_name)

            # Parse argument define
            argument_kwargs = kwargs_from_param(
                param,
                doc=param_to_description,
                class_or_function_name=class_or_function.__name__,
                ignored_flag=ignored_flag,
                defaults=defaults,
                prefix=prefix,
            )

            # Skip ignored args
            if ignored_flag:
                if argument_kwargs.get("required", False):
                    builder.partial = True
                continue

            # Add argument
            group.add_argument(f"--{prefix}{param_name}", **argument_kwargs)

            # Fill builder param names
            if dest:
                builder.keyword_arguments[param_name] = prefix + param_name

        if dest and builder.pass_extra_kwargs:
            trim_left = len(prefix)
            for prefixed_name, value in defaults.items():
                param_name = prefixed_name[trim_left:]
                if is_ignored(param_name): continue
                group.add_argument(f"--{prefixed_name}", **kwargs_from_value(value))
                builder.keyword_arguments[param_name] = prefixed_name


    def add_options(self, **kwargs):
        for argname, value in kwargs.items():
            self.add_argument(f"--{argname}", **kwargs_from_value(value))

    def parse_known_args(self, args=None, namespace=None):
        args, argv = super().parse_known_args(args=args, namespace=namespace)
        for dest, builder in self._typed_group_args.items():
            for param_name, prefixed_name in builder.keyword_arguments.items():
                builder.keyword_arguments[param_name] = getattr(args, prefixed_name)
            if not builder.partial:
                builder = builder()
            setattr(args, dest, builder)
        return args, argv


@dataclass
class OutputBuilder:
    class_or_function: Union[FunctionType, ClassType]
    partial: bool = False
    keyword_arguments: Dict = field(default_factory=dict)
    pass_extra_kwargs: bool = False

    def __call__(self, *args: Any, **kwds: Any) -> OutputType:
        return partial(self.class_or_function, **self.keyword_arguments)(*args, **kwds)

    def __repr__(self) -> str:
        params = [f"{k}={v}" for k, v in self.keyword_arguments.items()]
        if self.pass_extra_kwargs:
            params.append("**kwargs")
        params = ', '.join(params)
        return f"<partial {self.class_or_function.__name__}({params})>"


def kwargs_from_param(
    param: Parameter,
    class_or_function_name: str,
    doc: Dict[str, str] = dict(),
    defaults=dict(),
    explicit_bool=True,
    ignored_flag=False,
    prefix="",
):
    kwargs = dict()

    prefixed_name = prefix + param.name

    # Get the default or required of the argument
    if prefixed_name in defaults:
        kwargs["default"] = defaults[prefixed_name]
        del defaults[prefixed_name]
    elif param.default != Parameter.empty:
        kwargs["default"] = param.default
    else:
        kwargs["required"] = True

    # Parse type of the argument
    if not ignored_flag:
        # Get type of the argument
        if param.annotation != Parameter.empty:
            # Any type defaults to str (needed for dataclasses where all non-default attributes must have a type)
            if param.annotation is Any:
                var_type = str
            # Otherwise, get the type of the argument
            else:
                var_type = param.annotation
        # Infer type by the default value
        elif not kwargs.get("required", False):
            var_type = type(kwargs["default"])
        else:
            raise ArgumentTypeError(
                f"The parameter named '{param.name}' of '{class_or_function_name}' has not defined type annotation or default value to infer its type. Please add either to it, or add the param name to the 'ignore_args'. "
            )

        # Unbox Union[type] (Optional[type]) and set var_type = type
        if get_origin(var_type) in OPTIONAL_TYPES:
            var_args = get_args(var_type)

            # If type is Union or Optional without inner types, set type to equivalent of Optional[str]
            if len(var_args) == 0:
                var_args = (str, type(None))

            if len(var_args) > 0:
                var_type = var_args[0]

                # If var_type is tuple as in Python 3.6, change to a typing type
                # (e.g., (typing.List, <class 'bool'>) ==> typing.List[bool])
                if isinstance(var_type, tuple):
                    var_type = var_type[0][var_type[1:]]

                explicit_bool = True

        # First check whether it is a literal type or a boxed literal type
        if is_literal_type(var_type):
            var_type, kwargs["choices"] = get_literals(var_type, param.name)

        elif (
            get_origin(var_type) in (List, list, Set, set)
            and len(get_args(var_type)) > 0
            and is_literal_type(get_args(var_type)[0])
        ):
            var_type, kwargs["choices"] = get_literals(
                get_args(var_type)[0], param.name
            )
            if kwargs.get("action") not in {"append", "append_const"}:
                kwargs["nargs"] = kwargs.get("nargs", "*")

        # Handle Tuple type (with type args) by extracting types of Tuple elements and enforcing them
        elif get_origin(var_type) in (Tuple, tuple) and len(get_args(var_type)) > 0:
            loop = False
            types = get_args(var_type)

            # Handle Tuple[type, ...]
            if len(types) == 2 and types[1] == Ellipsis:
                types = types[0:1]
                loop = True
                kwargs["nargs"] = "*"
            # Handle Tuple[()]
            elif len(types) == 1 and types[0] == tuple():
                types = [str]
                loop = True
                kwargs["nargs"] = "*"
            else:
                kwargs["nargs"] = len(types)

            # Handle Literal types
            types = [
                get_literals(tp, param.name)[0] if is_literal_type(tp) else tp
                for tp in types
            ]

            var_type = TupleTypeEnforcer(types=types, loop=loop)

        if get_origin(var_type) in BOXED_TYPES:
            # If List or Set or Tuple type, set nargs
            if get_origin(var_type) in BOXED_COLLECTION_TYPES and kwargs.get(
                "action"
            ) not in {
                "append",
                "append_const",
            }:
                kwargs["nargs"] = kwargs.get("nargs", "*")

            # Extract boxed type for Optional, List, Set
            arg_types = get_args(var_type)

            # Set defaults type to str for Type and Type[()]
            if len(arg_types) == 0 or arg_types[0] == EMPTY_TYPE:
                var_type = str
            else:
                var_type = arg_types[0]

            # Handle the cases of List[bool], Set[bool], Tuple[bool]
            if var_type == bool:
                var_type = boolean_type

        # If bool then set action, otherwise set type
        if var_type == bool:
            if explicit_bool:
                kwargs["type"] = boolean_type
                kwargs["choices"] = [
                    True,
                    False,
                ]  # this makes the help message more helpful
            else:
                action_cond = (
                    "true"
                    if kwargs.get("required", False) or not kwargs["default"]
                    else "false"
                )
                kwargs["action"] = kwargs.get("action", f"store_{action_cond}")
        elif kwargs.get("action") not in {"count", "append_const"}:
            kwargs["type"] = var_type

    # Set help
    if not ignored_flag:
        kwargs["help"] = "("

        # Type
        kwargs["help"] += type_to_str(var_type) + ", "

        # Required/default
        if kwargs.get("required", False):
            kwargs["help"] += "required"
        else:
            kwargs["help"] += f'default={repr(kwargs.get("default", None))}'

        kwargs["help"] += ")"

        # Description
        docstring_help = " " + doc.get(param.name, "")
        docstring_help = docstring_help.replace("Default", "docstring default")
        docstring_help = docstring_help.replace("DEFAULT", "docstring default")
        docstring_help = docstring_help.replace("default", "docstring default")
        kwargs["help"] += docstring_help

    return kwargs


def kwargs_from_value(
    value: Any
):
    kwargs = dict()
    kwargs["type"] = type(value)
    kwargs["default"] = value

    # Set help
    if True:
        kwargs["help"] = "("

        # Type
        kwargs["help"] += type_to_str(kwargs["type"]) + ", "

        # Required/default
        if kwargs.get("required", False):
            kwargs["help"] += "required"
        else:
            kwargs["help"] += f'default={repr(value)}'

        kwargs["help"] += ")"

    return kwargs
